import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split
import matplotlib.pyplot as plt
import joblib
from sklearn.metrics import mean_squared_error, r2_score
import time
import tensorflow as tf
from tensorflow.keras.layers import MultiHeadAttention, LayerNormalization, Dense, Dropout, Embedding


gpus = tf.config.experimental.list_physical_devices('GPU')
if gpus:
  try:
    for gpu in gpus:
      tf.config.experimental.set_memory_growth(gpu, True)
  except RuntimeError as e:
    # Memory growth must be set before GPUs have been initialized
    print(e)

def html_to_rgb(color_string):
    color_string = color_string.strip()
    if color_string[0] == '#':
        color_string = color_string[1:]

    if len(color_string) == 3:
        return tuple(int(color_string[i]*2, 16) for i in range(3))
    elif len(color_string) == 6:
        return tuple(int(color_string[i:i+2], 16) for i in range(0, 6, 2))
    else:
        raise ValueError(f"Invalid HTML color code: {color_string}")

def rgb_to_norm(rgb):
    return tuple(val / 255 for val in rgb)

# Transformer Encoder Layer
def transformer_encoder(inputs, head_size, num_heads, ff_dim, dropout=0):
    # Normalization and Attention
    x = tf.keras.layers.LayerNormalization(epsilon=1e-6)(inputs)
    x = tf.keras.layers.MultiHeadAttention(
        key_dim=head_size, num_heads=num_heads, dropout=dropout
    )(x, x)
    x = tf.keras.layers.Dropout(dropout)(x)
    res = x + inputs

    # Feed Forward Part
    x = tf.keras.layers.LayerNormalization(epsilon=1e-6)(res)
    x = tf.keras.layers.Conv1D(filters=ff_dim, kernel_size=1, activation='relu')(x)
    x = tf.keras.layers.Dropout(dropout)(x)
    x = tf.keras.layers.Conv1D(filters=inputs.shape[-1], kernel_size=1)(x)
    return x + res


def TransformerBlock(embed_size, num_heads, ff_dim, rate=0.1):
    # 一个基本的Transformer块
    inputs = tf.keras.Input(shape=(None, embed_size))
    # 注意力层
    attention_output = MultiHeadAttention(num_heads=num_heads, key_dim=embed_size)(inputs, inputs)
    # 残差连接后的规范化层
    attention_output = LayerNormalization(epsilon=1e-6)(inputs + attention_output)
    # 前馈网络
    ff_output = Dense(ff_dim, activation="relu")(attention_output)
    ff_output = Dropout(rate)(ff_output)
    # 第二个残差连接和规范化层
    outputs = LayerNormalization(epsilon=1e-6)(attention_output + ff_output)
    return tf.keras.Model(inputs=[inputs], outputs=outputs)

# 定义一个接受一维输入的简单Transformer模型
def create_transformer_model(input_shape, embed_size, num_heads, ff_dim, num_transformer_blocks, mlp_units, dropout_rate, output_units):
    inputs = tf.keras.Input(shape=(input_shape,))
    # 可能需要一个嵌入层或其他预处理层，这里暂时省略
    x = Dense(embed_size)(inputs)
    x = tf.keras.layers.Reshape((1, embed_size))(x)  # 将输入重塑为序列
    # 堆叠Transformer块
    for _ in range(num_transformer_blocks):
        x = TransformerBlock(embed_size, num_heads, ff_dim)(x)
    # 展平输出以通过最后的MLP层
    x = tf.keras.layers.Flatten()(x)
    for units in mlp_units:
        x = Dense(units, activation='relu')(x)
        x = Dropout(dropout_rate)(x)
    outputs = Dense(output_units)(x)
    return tf.keras.Model(inputs=inputs, outputs=outputs)

# 定义模型参数
embed_size = 32  # 嵌入层大小
num_heads = 2    # 注意力头数
ff_dim = 32      # 前馈网络内的层数
num_transformer_blocks = 1  # 使用的Transformer块数量
mlp_units = [128]  # 后续MLP层的单元数
dropout_rate = 0.1  # Dropout率
output_units = 1  # 输出单元数，对应于回归问题

# 加载已保存的模型和缩放器
hardness_model = tf.keras.models.load_model('hardness_prediction_model.h5')
nano_model = tf.keras.models.load_model('nano_hardness_prediction_model.h5')
hardness_scaler = joblib.load('scaler.pkl')
nano_scaler = joblib.load('scaler_nano.pkl')

# 生成100-2000的距离的两个（纳米和维氏）19000个数据集
distances = np.linspace(100, 1800, num=17000).reshape(-1, 1)

# 使用模型生成硬度预测
hardness_pred = hardness_model.predict(hardness_scaler.transform(distances))
nano_hardness_pred = nano_model.predict(nano_scaler.transform(distances))

# 创建新的数据集
new_data = pd.DataFrame({
    'Distance': distances.flatten(),
    'Nano Hardness': nano_hardness_pred.flatten(),
    'Vickers Hardness': hardness_pred.flatten()
})

# 保存数据集
new_data.to_csv('硬度对应.csv', index=False)

# 现在，纳米硬度作为特征，维氏硬度作为结果
X = new_data['Nano Hardness'].values.reshape(-1, 1)
y = new_data['Vickers Hardness'].values

# 划分数据集
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)
X_train, X_val, y_train, y_val = train_test_split(X_train, y_train, test_size=0.25, random_state=42)

new_model = create_transformer_model(
    input_shape=1,
    embed_size=embed_size,
    num_heads=num_heads,
    ff_dim=ff_dim,
    num_transformer_blocks=num_transformer_blocks,
    mlp_units=mlp_units,
    dropout_rate=dropout_rate,
    output_units=output_units
)

# 编译模型
optimizer = tf.keras.optimizers.Adam(learning_rate=0.004)
new_model.compile(optimizer=optimizer, loss='mse', metrics=['mae'])

# 定义早停回调和每N步打印的回调
early_stopping = tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=1000, restore_best_weights=True)

class PrintEveryN(tf.keras.callbacks.Callback):
    def __init__(self, n):
        self.n = n

    def on_epoch_end(self, epoch, logs=None):
        if epoch % self.n == 0:
            print('Epoch', epoch)
            print('loss: ', logs['loss'])
            print('val_loss: ', logs['val_loss'])

class PredictionPlotter(tf.keras.callbacks.Callback):
    def on_epoch_end(self, epoch, logs=None):
        if epoch % 100 == 0:
            # Predict on test set
            y_pred = self.model.predict(X_test)
            y_all_pred = self.model.predict(X)

            # Calculate R2 and RMSE
            r2 = r2_score(y_test, y_pred)
            rmse = np.sqrt(mean_squared_error(y_test, y_pred))

            # Plot actual vs predicted
            plt.figure(figsize=(7, 6))
            plt.scatter(y_test, y_pred, s=10, color=rgb_to_norm(html_to_rgb('#1597A5')),label='Actual vs Predicted',linewidths=3)
            plt.plot(y_test, y_test, color=rgb_to_norm(html_to_rgb('#FFC239')), label='Perfect Prediction', linewidth=3)
            plt.xlabel('Vickers Hardness Actual', fontsize=18)
            plt.ylabel('Vickers Hardness Predicted', fontsize=18)
            plt.title('Actual vs Predicted (Epoch: {})'.format(epoch), fontsize=20)
            plt.text(0.6, 0.3, 'R2: {:.4f}\nrmse: {:.2f}'.format(r2, rmse), transform=plt.gca().transAxes, fontsize=20,
                     verticalalignment='top')
            plt.legend(fontsize=18)
            plt.tick_params(labelsize=12)
            plt.savefig('E:\我的科研论文\利用神经网路通过纳米硬度预测维氏硬度\纳米-维氏\Transformer\Actual vs Predicted Epoch_{}.png'.format(epoch), dpi=300)
            plt.show()

            # Plot Nano Hardness vs Predicted Vickers Hardness
            plt.figure(figsize=(7, 6))
            plt.scatter(new_data['Nano Hardness'], new_data['Vickers Hardness'], color=rgb_to_norm(html_to_rgb('#1597A5')), label='Actual', linewidths=1)
            plt.plot(new_data['Nano Hardness'], y_all_pred, color=rgb_to_norm(html_to_rgb('#FFC239')), label='Predicted', linewidth=3)
            plt.xlabel('Nano Hardness', fontsize=18)
            plt.ylabel('Predicted Vickers Hardness', fontsize=18)
            plt.title('Predicted Hardness (Epoch: {})'.format(epoch), fontsize=20)
            plt.legend(fontsize=18)
            plt.tick_params(labelsize=12)
            plt.savefig('E:\我的科研论文\利用神经网路通过纳米硬度预测维氏硬度\纳米-维氏\Transformer\Predicted Hardness Epoch_{}.png'.format(epoch), dpi=300)
            plt.show()



prediction_plotter = PredictionPlotter()
# 将这个回调加入模型训练
print_every_100 = PrintEveryN(100)

#计时
start = time.time()
history = new_model.fit(X_train, y_train, validation_data=(X_val, y_val), batch_size=128, epochs=3000, callbacks=[early_stopping, print_every_100, prediction_plotter], verbose=0)
#计时
end = time.time()
print('训练时间：',end-start)

new_model.save('prediction_model.h5')
loaded_model = tf.keras.models.load_model('prediction_model.h5')

# 在测试集上评估模型
test_loss, test_mae = loaded_model.evaluate(X_test, y_test)
print(f"测试集上的平均绝对误差 (MAE): {test_mae}")

# 预测值
y_pred = loaded_model.predict(X_test)

# 画出实际值和预测值的偏差图
plt.figure(figsize=(10, 10))
plt.scatter(y_test, y_pred, s=10)
plt.xlabel('Vickers Hardness Actual')
plt.ylabel('Vickers Hardness Predicted')
plt.title('Actual vs Predicted')
plt.show()

# 画出纳米硬度对应的维氏硬度预测曲线
plt.figure(figsize=(10, 10))
plt.plot(new_data['Nano Hardness'], new_data['Vickers Hardness'])
plt.xlabel('Nano Hardness')
plt.ylabel('Predicted Vickers Hardness')
plt.title('Nano Hardness vs Predicted Vickers Hardness')
plt.show()

# 绘制训练和验证损失曲线（对数形式）
plt.figure(figsize=(12, 9))
plt.plot(history.history['loss'], label='Training loss', lw=4)
plt.plot(history.history['val_loss'], label='Validation loss', lw=4)
plt.yscale('log')  # 设置y轴为对数刻度
plt.title('Training and Validation Loss', fontsize=36)
plt.xlabel('Epochs', fontsize=36)
plt.ylabel('Loss (log scale)', fontsize=36)
plt.tick_params(labelsize=28)
plt.legend(fontsize=30)
plt.savefig('E:\我的科研论文\利用神经网路通过纳米硬度预测维氏硬度\纳米-维氏\Transformer\Training and Validation Loss (log scale).png', dpi=300)
plt.show()

# 11. 用户交互系统
while True:
    user_input = input("请输入一个渗层深度值（或输入'退出'以退出程序）：")

    if user_input.lower() == '退出':
        break

    try:
        input_value = float(user_input)
    except ValueError:
        print("输入无效，请输入一个数字。")
        continue

    # 将用户输入缩放并进行预测
    input_value = np.array([[input_value]])  # 对input_value进行包装
    prediction = loaded_model.predict(input_value)

    print(f"预测的硬度为：{prediction[0][0]}")



