import pandas as pd
import numpy as np
import tensorflow as tf
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt
import joblib
from sklearn.metrics import mean_squared_error, r2_score

gpus = tf.config.experimental.list_physical_devices('GPU')
if gpus:
  try:
    for gpu in gpus:
      tf.config.experimental.set_memory_growth(gpu, True)
  except RuntimeError as e:
    # Memory growth must be set before GPUs have been initialized
    print(e)

def html_to_rgb(color_string):
    color_string = color_string.strip()
    if color_string[0] == '#':
        color_string = color_string[1:]

    if len(color_string) == 3:
        return tuple(int(color_string[i]*2, 16) for i in range(3))
    elif len(color_string) == 6:
        return tuple(int(color_string[i:i+2], 16) for i in range(0, 6, 2))
    else:
        raise ValueError(f"Invalid HTML color code: {color_string}")

def rgb_to_norm(rgb):
    return tuple(val / 255 for val in rgb)

# 1. 读取CSV文件
data = pd.read_csv('硬度数据.csv')

# 2. 处理数据
# 假设第一列是渗层深度（输入特征），第二列是硬度（目标值）
X = data.iloc[:, 0].values.reshape(-1, 1)
y = data.iloc[:, 1].values

# 3. 划分数据集
X_temp, X_test, y_temp, y_test = train_test_split(X, y, test_size=0.2, random_state=42)
X_train, X_val, y_train, y_val = train_test_split(X_temp, y_temp, test_size=0.25, random_state=42)

# 4. 特征缩放
scaler = StandardScaler()
X_train = scaler.fit_transform(X_train)
X_val = scaler.transform(X_val)
X_test = scaler.transform(X_test)

# 保存缩放器
joblib.dump(scaler, 'scaler.pkl')



# 5. 建立神经网络模型
model = tf.keras.Sequential([
    tf.keras.layers.Dense(32, activation='sigmoid', input_shape=(1,)),
    tf.keras.layers.Dense(64, activation='sigmoid'),
    tf.keras.layers.Dense(128, activation='sigmoid'),  # 新增的全连接层
    tf.keras.layers.Dense(1)
])

# 6. 编译模型
optimizer = tf.keras.optimizers.Adam(learning_rate=0.01)
model.compile(optimizer=optimizer, loss='mse', metrics=['mae'])

# 7. 定义早停回调
#early_stopping = tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=1000, restore_best_weights=True)

# 8. 训练模型
class PrintEveryN(tf.keras.callbacks.Callback):
    def __init__(self, n):
        self.n = n

    def on_epoch_end(self, epoch, logs=None):
        if epoch % self.n == 0:
            print('Epoch', epoch)
            print('loss: ', logs['loss'])
            print('val_loss: ', logs['val_loss'])

X_all = np.concatenate([X_train, X_val, X_test])
y_all = np.concatenate([y_train, y_val, y_test])

# 对预测数据进行逆缩放
X_all_unscaled = scaler.inverse_transform(X_all)

class PredictionPlotter(tf.keras.callbacks.Callback):
    def on_epoch_end(self, epoch, logs=None):
        if epoch % 200 == 0:  # Check if it's the 200th epoch
            # Create a new x value array from 0 to the max value
            x_values = np.linspace(0, X_all_unscaled.max(), num=500).reshape(-1, 1)

            # Scale the new x value array
            x_values_scaled = scaler.transform(x_values)



            # Use the model to make predictions
            y_values_pred = self.model.predict(x_values_scaled)

            # Plot the actual data
            plt.scatter(X_all_unscaled, y_all, color=rgb_to_norm(html_to_rgb('#1597A5')), label='Actual data')

            # Plot the predicted data curve
            plt.plot(x_values, y_values_pred, color=rgb_to_norm(html_to_rgb('#FFC239')), label='Predicted data')

            plt.xlabel('Depth', fontsize=18)
            plt.ylabel('Hardness', fontsize=18)
            plt.title('Depth vs Hardness (Epoch: {})'.format(epoch), fontsize=20)
            plt.legend(fontsize=16)
            plt.tick_params(labelsize=12)
            plt.savefig('E:\我的科研论文\利用神经网路通过纳米硬度预测维氏硬度\维氏硬度图片\Depth-hardness_epoch_{}.png'.format(epoch))
            plt.show()

class CustomCallback(tf.keras.callbacks.Callback):
    def __init__(self, x_all, y_all):
        super().__init__()
        self.x_all = x_all
        self.y_all = y_all

    def on_epoch_end(self, epoch, logs=None):
        if epoch % 200 == 0:  # Check if it's the 200th epoch
            # Use the model to make predictions
            y_values_pred = self.model.predict(self.x_all)

            # Calculate R2 and RMSE
            r2 = r2_score(self.y_all, y_values_pred)
            rmse = np.sqrt(mean_squared_error(self.y_all, y_values_pred))

            # Plot the actual data
            plt.scatter(self.y_all, y_values_pred,  color=rgb_to_norm(html_to_rgb('#1597A5')), label='Predicted data')

            # Plot the predicted data
            plt.plot(self.y_all, self.y_all,color=rgb_to_norm(html_to_rgb('#FFC239')), label='Actual data')

            plt.xlabel('Actual Hardness', fontsize=18)
            plt.ylabel('Predicted Hardness', fontsize=18)
            plt.title('Depth vs Hardness (Epoch: {})'.format(epoch), fontsize=20)
            plt.text(0.6, 0.3, 'R2: {:.4f}\nrmse: {:.2f}'.format(r2, rmse), transform=plt.gca().transAxes, fontsize=20, verticalalignment='top')
            plt.legend(fontsize=16)
            plt.tick_params(labelsize=12)
            plt.savefig('E:\我的科研论文\利用神经网路通过纳米硬度预测维氏硬度\维氏硬度图片\A-P_hardness_epoch_{}.png'.format(epoch))
            plt.show()

X_all = np.concatenate([X_train, X_val, X_test])
y_all = np.concatenate([y_train, y_val, y_test])

custom_callback = CustomCallback(X_all, y_all)

# Create an instance of the custom callback
prediction_plotter = PredictionPlotter()

print_every_100 = PrintEveryN(100)
# 使用批次大小为100进行训练
history = model.fit(X_train, y_train, batch_size=100, epochs=6000, validation_data=(X_val, y_val),callbacks=[print_every_100, prediction_plotter, custom_callback], verbose=0)

# 9. 导出模型
model.save('hardness_prediction_model.h5')

# 10. 加载已保存的模型
loaded_model = tf.keras.models.load_model('hardness_prediction_model.h5')

# 11. 在测试集上评估模型
test_loss, test_mae = loaded_model.evaluate(X_test, y_test)
print(f"测试集上的平均绝对误差（MAE）：{test_mae}")

# 12. 绘制实际数据和预测数据的曲线
X_all = np.concatenate([X_train, X_val, X_test])
y_all = np.concatenate([y_train, y_val, y_test])

# 使用模型进行预测
y_pred = loaded_model.predict(X_all)

# 对预测数据进行逆缩放
X_all_unscaled = scaler.inverse_transform(X_all)

# 创建一个新的x值数组，从0到最大值
x_values = np.linspace(0, X_all_unscaled.max(), num=500).reshape(-1, 1)

# 缩放新的x值数组
x_values_scaled = scaler.transform(x_values)

# 使用模型进行预测
y_values_pred = loaded_model.predict(x_values_scaled)
plt.xlim([0, X_all_unscaled.max()])
# 绘制实际数据
plt.scatter(X_all_unscaled, y_all, color='blue', label='Actual data')

# 绘制预测数据曲线
plt.plot(x_values, y_values_pred, color='red', label='Predicted data')

plt.xlabel('Infiltration Depth')
plt.ylabel('Hardness')
plt.title('Infiltration Depth vs Hardness')
plt.legend()
plt.show()

# 11. 用户交互系统
while True:
    user_input = input("请输入一个渗层深度值（或输入'退出'以退出程序）：")

    if user_input.lower() == '退出':
        break

    try:
        input_value = float(user_input)
    except ValueError:
        print("输入无效，请输入一个数字。")
        continue

    # 将用户输入缩放并进行预测
    scaled_input = scaler.transform([[input_value]])
    prediction = loaded_model.predict(scaled_input)

    print(f"预测的硬度为：{prediction[0][0]}")
